#!/usr/bin/env python3

from Bio import SeqIO
from Bio.Seq import Seq
from Bio.SeqRecord import SeqRecord
import os
import argparse
from functools import reduce

def evaluateRBS(sequence, maxdist, minspace=3, maxspace=7, atg = 'ATG', rbs = 'AGGAGG', len_rbs=6):
    """ Evaluate whether a given sequence is a good RBS"""
    disttable = dict()
    relativefirstpostochange = dict()
    newsequence = dict()
    for len_spacer in range(minspace, maxspace+1):
        matchrbs = [sequence[i] == x for (i, x) in enumerate(rbs)]
        drbs = matchrbs.count(False)
        matchatg = [sequence[i+len_rbs+len_spacer] == x for (i, x) in enumerate(atg)]
        datg = matchatg.count(False)
        disttable[len_spacer] = (drbs+datg)
        if drbs > 0:
            relativefirstpostochange[len_spacer] = matchrbs.index(False)
            newsequence[len_spacer] = list(sequence[:len_rbs+len_spacer+3])
            newsequence[len_spacer][relativefirstpostochange[len_spacer]] = rbs[relativefirstpostochange[len_spacer]]
        elif datg > 0:
            relativefirstpostochange[len_spacer] = len_rbs + len_spacer+matchatg.index(False)
            newsequence[len_spacer] = list(sequence[:len_rbs+len_spacer+3])
            newsequence[len_spacer][relativefirstpostochange[len_spacer]] = atg[matchatg.index(False)]
        else:
            relativefirstpostochange[len_spacer] = -1
            newsequence[len_spacer] = list(sequence[:len_rbs+len_spacer+3])
    return [(len_spacer, "".join(newsequence[len_spacer]), disttable[len_spacer], relativefirstpostochange[len_spacer]) for len_spacer in list(disttable.keys()) if disttable[len_spacer] <= maxdist]


def treatRBS(origgenesequence, pos, rbssequence, len_spacer, nb_nonsyn, pos_nonsyn, differ, save, detaildir, removestart=2, removefs=True, optimize_beginning=True, len_rbs=6):
    """ Once a candidate RBS has been found, try to make it suitable for expression of a downstream protein
    This function reports all changes it makes in a text file
    These changes aim at: removing rare codons in both existing and new reading frames, removing stop codons downstream the new start in the new reading frame, removing the alternative start codons in the new reading frame, removing mono-nucleotide repeats (mutagenic motifs)
    """
    # In one text file per RBS, output the detail of every changed nucleotide
    if detaildir:
        rbsfile = open("%s/%d.txt" % (detaildir, pos), "w")
        print("-- Original sequence\n%s" % str(origgenesequence), file=rbsfile)
    else:
        rbsfile = None
    # Compute the new sequence from the original sequence and the RBS-START
    posfirstcds = pos+len_rbs+len_spacer+3
    assert len(rbssequence) == len_rbs+len_spacer+3  # posfirstcds-pos
    # We do not use rbssequence entirely because we do not need the end
    genesequence = origgenesequence[:pos] + rbssequence + origgenesequence[posfirstcds:]
    shift = posfirstcds % 3
    if shift == 0:
        # Then this RBS is not usefull because in frame
        return
    fr = posfirstcds-shift+3  # The position in the upstream gene where we will start considering making synonymous change to eliminate STOP codons in our downstream new frame
    # We do not start at posfirtscds-shift otherwise we would risk to modify the RBS we just created.
    if not (len(genesequence[fr:]) % 3 == 0):
        assert(len(origgenesequence) % 3 == len(genesequence) % 3)
        if rbsfile:
            print("Warning: the original sequence length is not a multiple of 3 (pseudogene?), cutting...", file=rbsfile)
        genesequence = genesequence[:len(genesequence) - len(genesequence) % 3]
        origgenesequence = origgenesequence[:len(origgenesequence) - len(origgenesequence) % 3]

    # Output the synonymous changes we made to create this RBS-START
    if rbsfile:
        print("-- Creation of the RBS:", file=rbsfile)
        for i in differ:
            print("  At position %d replaced %s by %s (synonymous) to create a perfect RBS-START"
                  % (i, origgenesequence[i], genesequence[i]), file=rbsfile)

    # Report the possible non synonymous change made to create a perfect RBS-START
    if nb_nonsyn == 1:
        posnonsyn = pos+pos_nonsyn
        origcodon = origgenesequence[posnonsyn:posnonsyn+3].translate()
        newcodon = genesequence[posnonsyn:posnonsyn+3].translate()
        if rbsfile:
            print("  At position %d replaced %s by %s (non synonymous, %s -> %s) to create a perfect RBS-START"
                  % (posnonsyn, origgenesequence[posnonsyn], genesequence[posnonsyn], origcodon, newcodon), file=rbsfile)

    # Make synonymous changes in existing gene to optimize expression by removing rare codons
    # In some cases, codon optimization is also a good way of removing cryptit promoters
    optimized_end_galK = readingframes.optimize(str(genesequence[fr+6:]))
    if optimize_beginning and (fr > 25):
        optimized_beg_galK = readingframes.optimize(str(genesequence[:fr-24]))
        optimized_genesequence = Seq(optimized_beg_galK)+genesequence[fr-24:]
        optimized_genesequence = optimized_genesequence[:fr+6]+Seq(optimized_end_galK)
    else:
        optimized_genesequence = genesequence[:fr+6]+Seq(optimized_end_galK)
    if rbsfile:
        print("-- Sequence after optimizing the frame of galK\n%s" % optimized_genesequence, file=rbsfile)

    # Remove the stop codons that could exist in our new frame (aph, after the ATG) without modifying what is coded by the existing gene in main frame (GalK)
    (end_genesequence_wostop, changedpositionstop, remaining_stops) = readingframes.removestopinframepx(optimized_genesequence[fr:], shift, False)

    # Remove the start codons that could exist in our new frame (aph) without modifying what is coded by the existing gene in main frame (GalK)
    if (removestart == 1):  # We only remove starts after the ATG of aph
        (end_genesequence_wostopstart, changedpositionstart, remaining_starts) = readingframes.removestartinframepx(str(end_genesequence_wostop), shift, False)
        modified_genesequence = optimized_genesequence[:fr]+Seq(end_genesequence_wostopstart)
    elif (removestart == 2):  # We also remove starts before the ATG of aph
        (end_genesequence_wostopstart, rel_changedpositionstart, rel_remaining_starts) = readingframes.removestartinframepx(str(end_genesequence_wostop), shift, False)
        (start_genesequence_wostart, abs_changedpositionstart, abs_remaining_starts) = readingframes.removestartinframepx(str(optimized_genesequence[:pos-pos % 3]), shift, False)
        modified_genesequence = Seq(start_genesequence_wostart) + optimized_genesequence[pos-pos % 3:fr] + Seq(end_genesequence_wostopstart)
        changedpositionstart = [x-fr for x in abs_changedpositionstart]+rel_changedpositionstart
        remaining_starts = [x-fr for x in abs_remaining_starts]+rel_remaining_starts
    else:  # We do not remove starts at all
        modified_genesequence = optimized_genesequence[:fr] + end_genesequence_wostop
        changedpositionstart = []
        remaining_starts = []

    if rbsfile:
        # Output information about the synonymous changes we made to remove the stop codons
        print('-- Remove stop codons in new frame', file=rbsfile)
        for p in changedpositionstop:
            # It is possible that one change is reverted by the next one if changing two close positions, in which case we do not need to report
            if str(optimized_genesequence[fr+p]) != str(modified_genesequence[fr+p]):
                print("  At position %d replaced %s by %s (synonymous) to eliminate a STOP codon"
                      % (fr+p, optimized_genesequence[fr+p], modified_genesequence[fr+p]), file=rbsfile)

        # Output information about potential remaining STOP codons we were not able to remove
        for r in remaining_stops:
            print("At position %d unable to remove a STOP codon" % (fr+r), file=rbsfile)

        # Output information about the synonymous changes we made to remove the start codons
        print("-- Remove start codons in new frame", file=rbsfile)
        for p in changedpositionstart:
            if str(optimized_genesequence[fr+p]) != str(modified_genesequence[fr+p]):
                print("  At position %d replaced %s by %s (synonymous) to eliminate a START codon"
                      % (fr+p, optimized_genesequence[fr+p], modified_genesequence[fr+p]), file=rbsfile)

        # Output information about potential remaining START codons we were not able to remove
        for r in remaining_starts:
            print("At position %s unable to remove a START codon" % (fr+r), file=rbsfile)

        if removefs:
            print("-- Remove FS hotspots (single nucleotide repeats)", file=rbsfile)
        # Remove FS hotspots (everywhere, not only in the overlapping sequence)
        (nofs_sequence, remaininghotspots) = readingframes.removeFShotspots2frame(str(modified_genesequence), shift, 4, pos, pos+len_rbs+len_spacer+3)
        if rbsfile:
            for pn in [i for (i, x) in enumerate(str(modified_genesequence)) if not x == nofs_sequence[i]]:
                if str(nofs_sequence[pn]) != str(modified_genesequence[pn]):
                    print("  At position %d replaced %s by %s (synonymous) to eliminate a FS hotspot"
                          % (pn, modified_genesequence[pn], nofs_sequence[pn]), file=rbsfile)

            # Output information about potential FS hotspots we were not able to remove
            for pn in remaininghotspots:
                print("At position %d unable to remove a FS hotspot" % pn, file=rbsfile)

        modified_genesequence = Seq(nofs_sequence)

    # Remove rare codons in alternative frame
    (end_genesequence_worare, changedpositionrare, remaining_rare) = readingframes.removerarecodonsinframepx(str(modified_genesequence)[fr:], shift, 4, 8.0, False)
    modified_genesequence = modified_genesequence[:fr] + Seq(end_genesequence_worare)
    if rbsfile:
        # Output information about changes we made to remove these codons
        print("-- Remove rare codons", file=rbsfile)
        for p in changedpositionrare:
            if str(optimized_genesequence[fr+p]) != str(modified_genesequence[fr+p]):
                print("  At position %d replaced %s by %s (synonymous) to eliminate a rare codon"
                      % (fr+p, optimized_genesequence[fr+p], modified_genesequence[fr+p]), file=rbsfile)
        for r in remaining_rare:
            print("At position %d unable to remove a rare codon" % (fr+r), file=rbsfile)

    if rbsfile:
        print("-- Final sequence\n%s" % modified_genesequence, file=rbsfile)
        rbsfile.close()

    # Compute other information we want about this overlap (number of syn/nonsyn changes, absolute and relative size, ...)
    distAA = [str(modified_genesequence.translate())[i] == x for (i, x) in enumerate(str(origgenesequence.translate()))].count(False)
    distBP = [modified_genesequence[i] == x for (i, x) in enumerate(origgenesequence)].count(False)
    sizeoverlap = (len(genesequence)-(pos+len_rbs+len_spacer))
    proportion = float(sizeoverlap)/len(genesequence)
    if rbsfile:
        print('%d,%d,%.18f,%d,%d,%d,%d,%d,%s,%s' % (pos, sizeoverlap, proportion, shift, len_spacer, distBP, distAA, len(remaining_stops), 'ff266af', str(modified_genesequence)), file=save)
    else:  # Quiet mode, do not output the full modified sequence
        print('%d,%d,%.18f,%d,%d,%d,%d,%d,%s' % (pos, sizeoverlap, proportion, shift, len_spacer, distBP, distAA, len(remaining_stops), 'ff266af'), file=save)


def findRBS(genesequence, save, detaildir=None):
    """ Try to create RBS in a coding sequence making only synonymous changes
    This is achieved by local bruteforce: at each position, we consider all possible synonymous substitutions in the next 18 nucleotides
    """
    lseq = len(genesequence)
    local = 18  # RBS (6) + spacer (<=7) + ATG (3) rounded to the next codon
    for pos in range(0, lseq-local):  # For each position in the gene, we try to start an RBS there
        totreat = dict()
        shift = pos % 3
        # We look at the 18 next nucleotides
        candidate = genesequence[pos-shift:pos-shift+18]
        assert (len(candidate) == 18)
        # We transform the list of 18 nucleotides into a list of 6 codons
        codonliste = [str(candidate[k:k+3]) for k in range(0, local, 3)]
        # We look at each possible combination of synonymous changes
        for (synonymous_combination, rarity_score) in readingframes.generate_synonymous_combinations(codonliste):
            syn_candidate = Seq(reduce(lambda x, y: x+y, synonymous_combination))
            differ = set([i+pos-shift for (i, x) in enumerate(candidate) if syn_candidate[i] != x])
            rbs_candidate = syn_candidate[shift:]
            nbps = [candidate[i] == x for (i, x) in enumerate(syn_candidate)].count(False)
            assert (len(differ) == nbps)
            # We calculate the score of this candidate. We have to be carreful that it can happens it is a good candidate for several different RBS configurations (different spacer sizes), so we want to find all the possible configurations
            # We allow RBS+spacer+ATG that have one nucleotide different from the consensus (can be turned into consensus making one non synonymous change)
            allconfigs = evaluateRBS(str(rbs_candidate), 1)
            # nb_nonsyn is 0 or 1, so pos_nonsyn is only one number
            for (len_spacer, mod_rbs_candidate, nb_nonsyn, pos_nonsyn) in allconfigs:
                if (
                    mod_rbs_candidate not in totreat
                    or totreat[mod_rbs_candidate][1] > nb_nonsyn
                    or ((totreat[mod_rbs_candidate][1] == nb_nonsyn) and (totreat[mod_rbs_candidate][3] > nbps))
                ):
                    totreat[mod_rbs_candidate] = (len_spacer, nb_nonsyn, pos_nonsyn, nbps, differ, rarity_score)
                    # TODO: maybe index by rbs_candidate and not mod_rbs_candidate
        # Treat the best RBS we found
        if totreat:
            besttotreat = sorted(totreat, key=lambda x: totreat[x][1]*1000 + int(totreat[x][5]*30) + totreat[x][3])[0]
            treatRBS(genesequence, pos, besttotreat, totreat[besttotreat][0], totreat[besttotreat][1], totreat[besttotreat][2], totreat[besttotreat][4], save, detaildir, removestart=2, removefs=True)


def main():
    parser = argparse.ArgumentParser(description="The RiBoSor creates alternative reading frames in an existing coding sequence, using genetic code redundancy")
    parser.add_argument('inputfile', type=str, help='input fasta file (containing one or several coding sequences)')
    parser.add_argument('--outputdir', type=str, help='output directory (will be inferred from the fasta file if not specified)', metavar="dir")
    parser.add_argument('--summary', action='store_true', default=False, help='only produce summary csv file (no detail by candidate), useful for screening')
    args = parser.parse_args()

    if not ".fasta" in args.inputfile:
        print("Warning: input should be a fasta file")

    if args.outputdir:
        outputdir = args.outputdir
    else:
        outputdir = os.path.splitext(args.inputfile)[0]
    os.makedirs(outputdir, exist_ok=True)

    records = list(SeqIO.parse(open(args.inputfile, "r"), "fasta"))
    for nseq, fasta_record in enumerate(records):
        bioseq = fasta_record.seq
        if len(records) > 1:
            seqname = fasta_record.name
        else:
            seqname = os.path.splitext(args.inputfile)[0]
        seqdir = os.path.join(outputdir, seqname)+'.details'
        if args.summary:
            seqdir = None
        else:
            os.mkdir(seqdir)
        outputfilename = os.path.join(outputdir,seqname)+'.csv'
        outputfile = open(outputfilename, "x")
        print("Start position, Length of the Overlap, Fraction protected, Frame, Spacer, Number of base pair changed, Number of amino acid changed, Number of remaining STOPS, Algorithm, New sequence", file=outputfile)
        try:
            findRBS(bioseq, outputfile, seqdir)
        except:
            print("Error while processing record %d of file %s" % (nseq, args.inputfile))

if __name__ == '__main__':
    import readingframes 
    #allsyn = readingframes.SynonymousCodons
    #allnonsyn = readingframes.NonSynonymousCodons
    main()
else:
    from ribosor import readingframes 
    #allsyn = readingframes.SynonymousCodons
    #allnonsyn = readingframes.NonSynonymousCodons
# This is rather ugly, but this permits to directly download and use this program as a script instead of installing it as a package



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="RiBoSor",
    url="https://github.com/afrenoy/RiBoSor",
    version="1.0",
    author="Antoine Frenoy",
    author_email="antoine.frenoy@pasteur.fr",
    description="Creates overlapping reading frames within an existing gene",
    long_description=long_description,
    long_description_content_type="text/markdown",
    classifiers=["Programming Language :: Python :: 3", "License :: OSI Approved :: GNU General Public License v3 (GPLv3)", "Topic :: Scientific/Engineering :: Bio-Informatics"],
    install_requires=["biopython"],
    entry_points={'console_scripts': ["ribosor = ribosor.ribosor:main"]},
    packages=["ribosor"],
    include_package_data=True,
    )



""" Functions to remove start codons, stop codons, mono-nucleotide repeats and rare codons, using genetic code redundancy
These functions can be used in the main reading or in an alternative reading frame
"""

from Bio import SeqIO
from Bio.Seq import Seq
from functools import reduce

""" Data and simple functions to work with the genetic code
"""


NonSynonymousCodons = {
    'TTC': ['TAT', 'TAC', 'TGG', 'TGG', 'ATC', 'ATA', 'ATT', 'TTG', 'TTA', 'CTC', 'CTG', 'CTA', 'CTT'],
    'TTT': ['TAT', 'TAC', 'TGG', 'TGG', 'ATC', 'ATA', 'ATT', 'TTG', 'TTA', 'CTC', 'CTG', 'CTA', 'CTT'],
    'TTG': ['ATC', 'ATA', 'ATT', 'ATG', 'GTC', 'GTA', 'GTG', 'GTT', 'TTT', 'TTC'],
    'TTA': ['ATC', 'ATA', 'ATT', 'ATG', 'GTC', 'GTA', 'GTG', 'GTT', 'TTT', 'TTC'],
    'CTT': ['ATC', 'ATA', 'ATT', 'ATG', 'GTC', 'GTA', 'GTG', 'GTT', 'TTT', 'TTC'],
    'CTC': ['ATC', 'ATA', 'ATT', 'ATG', 'GTC', 'GTA', 'GTG', 'GTT', 'TTT', 'TTC'],
    'CTA': ['ATC', 'ATA', 'ATT', 'ATG', 'GTC', 'GTA', 'GTG', 'GTT', 'TTT', 'TTC'],
    'CTG': ['ATC', 'ATA', 'ATT', 'ATG', 'GTC', 'GTA', 'GTG', 'GTT', 'TTT', 'TTC'],
    'ATT': ['GTC', 'GTA', 'GTG', 'GTT', 'TTG', 'TTA', 'CTC', 'CTG', 'CTA', 'CTT', 'TTT', 'TTC', 'TGT', 'TGC'],
    'ATC': ['GTC', 'GTA', 'GTG', 'GTT', 'TTG', 'TTA', 'CTC', 'CTG', 'CTA', 'CTT', 'TTT', 'TTC', 'TGT', 'TGC'],
    'ATA': ['GTC', 'GTA', 'GTG', 'GTT', 'TTG', 'TTA', 'CTC', 'CTG', 'CTA', 'CTT', 'TTT', 'TTC', 'TGT', 'TGC'],
    'ATG': ['GTC', 'GTA', 'GTG', 'GTT', 'TTG', 'TTA', 'CTC', 'CTG', 'CTA', 'CTT', 'TTT', 'TTC', 'TGT', 'TGC'],
    'GTT': ['ATC', 'ATA', 'ATT', 'TTG', 'TTA', 'CTC', 'CTG', 'CTA', 'CTT', 'ATG', 'GCT', 'GCC', 'GCA', 'GCG'],
    'GTC': ['ATC', 'ATA', 'ATT', 'TTG', 'TTA', 'CTC', 'CTG', 'CTA', 'CTT', 'ATG', 'GCT', 'GCC', 'GCA', 'GCG'],
    'GTA': ['ATC', 'ATA', 'ATT', 'TTG', 'TTA', 'CTC', 'CTG', 'CTA', 'CTT', 'ATG', 'GCT', 'GCC', 'GCA', 'GCG'],
    'GTG': ['ATC', 'ATA', 'ATT', 'TTG', 'TTA', 'CTC', 'CTG', 'CTA', 'CTT', 'ATG', 'GCT', 'GCC', 'GCA', 'GCG'],
    'TCT': ['ACC', 'ACT', 'ACA', 'ACG', 'GCT', 'GCC', 'GCA', 'GCG', 'AAC', 'AAT', 'GAC', 'GAT'],
    'TCC': ['ACC', 'ACT', 'ACA', 'ACG', 'GCT', 'GCC', 'GCA', 'GCG', 'AAC', 'AAT', 'GAC', 'GAT'],
    'TCA': ['ACC', 'ACT', 'ACA', 'ACG', 'GCT', 'GCC', 'GCA', 'GCG', 'AAC', 'AAT', 'GAC', 'GAT'],
    'TCG': ['ACC', 'ACT', 'ACA', 'ACG', 'GCT', 'GCC', 'GCA', 'GCG', 'AAC', 'AAT', 'GAC', 'GAT'],
    'CCT': ['GCT', 'GCC', 'GCA', 'GCG', 'GAC', 'GAT', 'CAG', 'CAA', 'GAA', 'GAG'],
    'CCC': ['GCT', 'GCC', 'GCA', 'GCG', 'GAC', 'GAT', 'CAG', 'CAA', 'GAA', 'GAG'],
    'CCA': ['GCT', 'GCC', 'GCA', 'GCG', 'GAC', 'GAT', 'CAG', 'CAA', 'GAA', 'GAG'],
    'CCG': ['GCT', 'GCC', 'GCA', 'GCG', 'GAC', 'GAT', 'CAG', 'CAA', 'GAA', 'GAG'],
    'ACT': ['TCT', 'TCA', 'TCG', 'TCC', 'AGT', 'AGG', 'GTC', 'GTA', 'GTG', 'GTT', 'GAA', 'GAG', 'GGG', 'GGC', 'GGA', 'GGT'],
    'ACC': ['TCT', 'TCA', 'TCG', 'TCC', 'AGT', 'AGG', 'GTC', 'GTA', 'GTG', 'GTT', 'GAA', 'GAG', 'GGG', 'GGC', 'GGA', 'GGT'],
    'ACA': ['TCT', 'TCA', 'TCG', 'TCC', 'AGT', 'AGG', 'GTC', 'GTA', 'GTG', 'GTT', 'GAA', 'GAG', 'GGG', 'GGC', 'GGA', 'GGT'],
    'ACG': ['TCT', 'TCA', 'TCG', 'TCC', 'AGT', 'AGG', 'GTC', 'GTA', 'GTG', 'GTT', 'GAA', 'GAG', 'GGG', 'GGC', 'GGA', 'GGT'],
    'GCT': ['GTC', 'GTA', 'GTG', 'GTT', 'GGG', 'GGC', 'GGA', 'GGT', 'TGT', 'TGC', 'ACC', 'ACT', 'ACA', 'ACG'],
    'GCC': ['GTC', 'GTA', 'GTG', 'GTT', 'GGG', 'GGC', 'GGA', 'GGT', 'TGT', 'TGC', 'ACC', 'ACT', 'ACA', 'ACG'],
    'GCA': ['GTC', 'GTA', 'GTG', 'GTT', 'GGG', 'GGC', 'GGA', 'GGT', 'TGT', 'TGC', 'ACC', 'ACT', 'ACA', 'ACG'],
    'GCG': ['GTC', 'GTA', 'GTG', 'GTT', 'GGG', 'GGC', 'GGA', 'GGT', 'TGT', 'TGC', 'ACC', 'ACT', 'ACA', 'ACG'],
    'TAT': ['TGG', 'TGG', 'ATG', 'TTG', 'TTA', 'CTC', 'CTG', 'CTA', 'CTT', 'ATC', 'ATA', 'ATT'],
    'TAC': ['TGG', 'TGG', 'ATG', 'TTG', 'TTA', 'CTC', 'CTG', 'CTA', 'CTT', 'ATC', 'ATA', 'ATT'],
    'CAT': ['TAT', 'TAC', 'AGG', 'CGC', 'CGA', 'CGT', 'AGA', 'CGC', 'AAC', 'AAT', 'GAA', 'GAG'],
    'CAC': ['TAT', 'TAC', 'AGG', 'CGC', 'CGA', 'CGT', 'AGA', 'CGC', 'AAC', 'AAT', 'GAA', 'GAG'],
    'CAA': ['AGG', 'CGC', 'CGA', 'CGT', 'AGA', 'CGC', 'AAC', 'AAT', 'GAC', 'GAT', 'CAT', 'CAC'],
    'CAG': ['AGG', 'CGC', 'CGA', 'CGT', 'AGA', 'CGC', 'AAC', 'AAT', 'GAC', 'GAT', 'CAT', 'CAC'],
    'AAT': ['TCT', 'TCA', 'TCG', 'TCC', 'AGT', 'AGG', 'CAT', 'CAC', 'GAA', 'GAG', 'GGG', 'GGC', 'GGA', 'GGT'],
    'AAC': ['TCT', 'TCA', 'TCG', 'TCC', 'AGT', 'AGG', 'CAT', 'CAC', 'GAA', 'GAG', 'GGG', 'GGC', 'GGA', 'GGT'],
    'AAA': ['AGG', 'CGC', 'CGA', 'CGT', 'AGA', 'CGC', 'GAA', 'GAG', 'CAG', 'CAA', 'TCT', 'TCA', 'TCG', 'TCC', 'AGT', 'AGG'],
    'AAG': ['AGG', 'CGC', 'CGA', 'CGT', 'AGA', 'CGC', 'GAA', 'GAG', 'CAG', 'CAA', 'TCT', 'TCA', 'TCG', 'TCC', 'AGT', 'AGG'],
    'GAT': ['AAC', 'AAT', 'GAA', 'GAG', 'TCT', 'TCA', 'TCG', 'TCC', 'AGT', 'AGG', 'CAG', 'CAA'],
    'GAC': ['AAC', 'AAT', 'GAA', 'GAG', 'TCT', 'TCA', 'TCG', 'TCC', 'AGT', 'AGG', 'CAG', 'CAA'],
    'GAA': ['AAA', 'AAG', 'GCT', 'GCC', 'GCA', 'GCG', 'AAC', 'AAT', 'TCT', 'TCA', 'TCG', 'TCC', 'AGT', 'AGG'],
    'GAG': ['AAA', 'AAG', 'GCT', 'GCC', 'GCA', 'GCG', 'AAC', 'AAT', 'TCT', 'TCA', 'TCG', 'TCC', 'AGT', 'AGG'],
    'TGT': ['GCT', 'GCC', 'GCA', 'GCG', 'ATC', 'ATA', 'ATT', 'TTG', 'TTA', 'CTC', 'CTG', 'CTA', 'CTT', 'TTT', 'TTC'],
    'TGC': ['GCT', 'GCC', 'GCA', 'GCG', 'ATC', 'ATA', 'ATT', 'TTG', 'TTA', 'CTC', 'CTG', 'CTA', 'CTT', 'TTT', 'TTC'],
    'TGG': ['TAT', 'TAC', 'TTT', 'TTC', 'ACC', 'ACT', 'ACA', 'ACG', 'CAT', 'CAC'],
    'CGT': ['AAA', 'AAG', 'CAG', 'CAA', 'CAT', 'CAC', 'GAA', 'GAG'],
    'CGC': ['AAA', 'AAG', 'CAG', 'CAA', 'CAT', 'CAC', 'GAA', 'GAG'],
    'CGA': ['AAA', 'AAG', 'CAG', 'CAA', 'CAT', 'CAC', 'GAA', 'GAG'],
    'CGG': ['AAA', 'AAG', 'CAG', 'CAA', 'CAT', 'CAC', 'GAA', 'GAG'],
    'AGA': ['AAA', 'AAG', 'CAG', 'CAA', 'CAT', 'CAC', 'GAA', 'GAG'],
    'AGG': ['AAA', 'AAG', 'CAG', 'CAA', 'CAT', 'CAC', 'GAA', 'GAG'],
    'AGT': ['ACC', 'ACT', 'ACA', 'ACG', 'GCT', 'GCC', 'GCA', 'GCG', 'AAC', 'AAT', 'GAC', 'GAT'],
    'AGG': ['ACC', 'ACT', 'ACA', 'ACG', 'GCT', 'GCC', 'GCA', 'GCG', 'AAC', 'AAT', 'GAC', 'GAT'],
    'GGT': ['GCT', 'GCC', 'GCA', 'GCG', 'AAC', 'AAT', 'TCT', 'TCA', 'TCG', 'TCC', 'AGT', 'AGG', 'GAC', 'GAT'],
    'GGC': ['GCT', 'GCC', 'GCA', 'GCG', 'AAC', 'AAT', 'TCT', 'TCA', 'TCG', 'TCC', 'AGT', 'AGG', 'GAC', 'GAT'],
    'GGA': ['GCT', 'GCC', 'GCA', 'GCG', 'AAC', 'AAT', 'TCT', 'TCA', 'TCG', 'TCC', 'AGT', 'AGG', 'GAC', 'GAT'],
    'GGG': ['GCT', 'GCC', 'GCA', 'GCG', 'AAC', 'AAT', 'TCT', 'TCA', 'TCG', 'TCC', 'AGT', 'AGG', 'GAC', 'GAT'],
    'ATG': ['TTG', 'TTA', 'CTC', 'CTG', 'CTA', 'CTT', 'GTC', 'GTA', 'GTG', 'GTT', 'ATC', 'ATA', 'ATT', 'CAG', 'CAA']
}


SynonymousCodons = {
    'TGT': ['TGC'],
    'TGC': ['TGT'],
    'GAC': ['GAT'],
    'GAT': ['GAC'],
    'TCT': ['TCG', 'TCA', 'TCC', 'AGC', 'AGT'],
    'TCG': ['TCT', 'TCA', 'TCC', 'AGC', 'AGT'],
    'TCA': ['TCT', 'TCG', 'TCC', 'AGC', 'AGT'],
    'TCC': ['TCT', 'TCG', 'TCA', 'AGC', 'AGT'],
    'AGC': ['TCT', 'TCG', 'TCA', 'TCC', 'AGT'],
    'AGT': ['TCT', 'TCG', 'TCA', 'TCC', 'AGC'],
    'CAA': ['CAG'],
    'CAG': ['CAA'],
    'ATG': [],
    'AAC': ['AAT'],
    'AAT': ['AAC'],
    'CCT': ['CCG', 'CCA', 'CCC'],
    'CCG': ['CCT', 'CCA', 'CCC'],
    'CCA': ['CCT', 'CCG', 'CCC'],
    'CCC': ['CCT', 'CCG', 'CCA'],
    'AAG': ['AAA'],
    'AAA': ['AAG'],
    'TAG': ['TGA', 'TAA'],
    'TGA': ['TAG', 'TAA'],
    'TAA': ['TAG', 'TGA'],
    'ACC': ['ACA', 'ACG', 'ACT'],
    'ACA': ['ACC', 'ACG', 'ACT'],
    'ACG': ['ACC', 'ACA', 'ACT'],
    'ACT': ['ACC', 'ACA', 'ACG'],
    'TTC': ['TTT'],
    'TTT': ['TTC'],
    'GCA': ['GCC', 'GCG', 'GCT'],
    'GCC': ['GCA', 'GCG', 'GCT'],
    'GCG': ['GCA', 'GCC', 'GCT'],
    'GCT': ['GCA', 'GCC', 'GCG'],
    'GGT': ['GGC', 'GGG', 'GGA'],
    'GGG': ['GGC', 'GGT', 'GGA'],
    'GGA': ['GGC', 'GGT', 'GGG'],
    'GGC': ['GGT', 'GGG', 'GGA'],
    'ATC': ['ATT', 'ATA'],
    'ATA': ['ATC', 'ATT'],
    'ATT': ['ATC', 'ATA'],
    'TTA': ['TTG', 'CTC', 'CTT', 'CTG', 'CTA'],
    'TTG': ['TTA', 'CTC', 'CTT', 'CTG', 'CTA'],
    'CTC': ['TTA', 'TTG', 'CTT', 'CTG', 'CTA'],
    'CTT': ['TTA', 'TTG', 'CTC', 'CTG', 'CTA'],
    'CTG': ['TTA', 'TTG', 'CTC', 'CTT', 'CTA'],
    'CTA': ['TTA', 'TTG', 'CTC', 'CTT', 'CTG'],
    'CAC': ['CAT'],
    'CAT': ['CAC'],
    'CGA': ['CGT', 'CGC', 'CGG', 'AGA', 'AGG'],
    'CGC': ['CGT', 'CGG', 'CGA', 'AGA', 'AGG'],
    'CGG': ['CGT', 'CGC', 'CGA', 'AGA', 'AGG'],
    'CGT': ['CGC', 'CGG', 'CGA', 'AGA', 'AGG'],
    'AGG': ['CGT', 'CGC', 'CGG', 'CGA', 'AGA'],
    'AGA': ['CGT', 'CGC', 'CGG', 'CGA', 'AGG'],
    'TGG': [],
    'GTA': ['GTC', 'GTG', 'GTT'],
    'GTC': ['GTA', 'GTG', 'GTT'],
    'GTG': ['GTA', 'GTC', 'GTT'],
    'GTT': ['GTA', 'GTC', 'GTG'],
    'GAG': ['GAA'],
    'GAA': ['GAG'],
    'TAC': ['TAT'],
    'TAT': ['TAC']
}


CodonUsage = {
    'GGG': 10.99,
    'GGA': 7.92,
    'GGT': 24.85,
    'GGC': 29.44,
    'GAG': 17.79,
    'GAA': 39.58,
    'GAT': 32.22,
    'GAC': 19.05,
    'GTG': 26.25,
    'GTA': 10.88,
    'GTT': 18.38,
    'GTC': 15.22,
    'GCG': 33.66,
    'GCA': 20.28,
    'GCT': 15.34,
    'GCC': 25.51,
    'AGG': 1.22,
    'AGA': 2.05,
    'CGG': 5.38,
    'CGA': 3.53,
    'CGT': 21.02,
    'CGC': 22.02,
    'AAG': 10.21,
    'AAA': 33.62,
    'AAT': 17.62,
    'AAC': 21.67,
    'ATG': 27.77,
    'ATA': 4.28,
    'ATT': 30.40,
    'ATC': 25.00,
    'ACG': 14.37,
    'ACA': 7.02,
    'ACT': 8.92,
    'ACC': 23.38,
    'TGG': 15.31,
    'TGT': 5.18,
    'TGC': 6.44,
    'TAG': 0.23,
    'TAA': 2.02,
    'TGA': 0.90,
    'TAT': 16.32,
    'TAC': 12.27,
    'TTT': 22.40,
    'TTC': 16.59,
    'AGT': 8.71,
    'AGC': 16.03,
    'TCG': 8.92,
    'TCA': 7.13,
    'TCT': 8.50,
    'TCC': 8.59,
    'CAG': 28.84,
    'CAA': 15.45,
    'CAT': 12.90,
    'CAC': 9.72,
    'TTG': 13.72,
    'TTA': 13.89,
    'CTG': 52.82,
    'CTA': 3.85,
    'CTT': 11.04,
    'CTC': 11.04,
    'CCG': 23.27,
    'CCA': 8.52,
    'CCT': 7.04,
    'CCC': 5.52,
}


def compute_rarity_score(x):
    """Compute the rarity score of a list of codons
    The higer this score, the more rare codons are used
    """
    return sum([1 + (8. - CodonUsage[c]) / 8. for c in x if (CodonUsage[c] <= 8.)])


def generate_synonymous_combinations(codons_list):
    """Try all possible combinations of sets of codons,
    taking one codon of each set,
    and sort them according to the rarity score
    with the combinantion using the less rare codons first
    """
    import itertools
    score = dict()
    for combination in itertools.product(*[SynonymousCodons[i] + [i] for i in codons_list]):
        score[combination] = compute_rarity_score(combination)
    return sorted(score.items(), key=lambda x: x[1])


mfsc = dict()  # Most Frequent Synonymous Codon
for codon in CodonUsage:
    if len(SynonymousCodons[codon]) == 0:
        mfsc[codon] = codon
        continue
    candidate = max(SynonymousCodons[codon], key=lambda sc: CodonUsage[sc])
    if CodonUsage[candidate] < 8.0 and CodonUsage[codon] > CodonUsage[candidate]:
        mfsc[codon] = codon
    else:
        mfsc[codon] = candidate


def optimize(s0):
    """Codon optimization: replace each codon by the most common alternative"""
    newseq = list(s0)
    for i in range(0, len(s0), 3):
        newcodon = mfsc[s0[i:i + 3]]
        newseq[i:i + 3] = newcodon
    s1 = ''.join(newseq)
    return s1


def removestopinframepx(s0, x, verbose=True):
    """Take a sequence and remove stop codons in alternative frame with only synonymous changes in main frame
    The sequence is taken as a BioSeq object and not as a python string
    Returns the modified sequence and the number of remaining stops
    """

    assert (x > 0) and (x < 3)
    prot0 = s0.translate()
    l0 = len(s0)

    sx = s0[x:l0 - 3 + x]
    tx = sx.translate()
    pt = tx.find('*')

    bummer = False

    changedposition = []

    # While there remains some stop codons
    while(pt > -1):
        # Find the involved codons in s0
        p1 = pt * 3
        c1 = str(s0[p1:p1 + 3])
        c2 = str(s0[p1 + 3:p1 + 6])
        # Try to change them with synonymous
        candfound = False
        allcand = dict()
        for ((cand1, cand2), rarity_score) in generate_synonymous_combinations([c1, c2]):
            new = list(s0)
            new[p1:p1 + 3] = cand1[0:3]
            new[p1 + 3:p1 + 6] = cand2[0:3]
            news0 = Seq("".join(new))
            # Check whether the stop codon has been removed in new sequence
            newsx = news0[x:l0 - 3 + x]
            newtx = newsx.translate()
            newpt = newtx.find('*', pt)
            if (newpt > pt) or (newpt == -1):
                candfound = True
                # Calculate the number of BPS
                allcand[(cand1, cand2)] = ([(c1 + c2)[i] == BP for (i, BP) in enumerate(cand1 + cand2)].count(False), rarity_score)
        if candfound:  # We found at least one candidate
            # Find the best one (less BPS and less rare codons) among all possible candidates
            (cand1, cand2) = sorted(allcand, key=lambda u: float(allcand[u][0]) + allcand[u][1] * 10.)[0]
            changedposition = changedposition + [p1 + i for (i, BP) in enumerate(cand1 + cand2) if BP != (c1 + c2)[i]]
            new = list(s0)
            new[p1:p1 + 3] = cand1[0:3]
            new[p1 + 3:p1 + 6] = cand2[0:3]
            news0 = Seq("".join(new))
            newsx = news0[x:l0 - 3 + x]
            newtx = newsx.translate()
            newpt = newtx.find('*', pt)
            assert(len(s0) == len(news0))
            if verbose:
                print("We removed stop codon at position " + str(pt * 3 + x) + ". ")
            s0 = news0
            sx = newsx
            tx = newtx
            pt = newpt
        else:
            bummer = True
            if verbose:
                print("Bummer ! We are unable to remove stop codon at position " + str(pt * 3 + x) + ". Let's skip this one and go on.")
            pt = tx.find('*', pt + 1)
            continue
    # We have removed all stop codons in frame x
    assert (str(s0.translate()) == str(prot0))
    assert (bummer or ('*' not in tx))
    return (s0, changedposition, [i * 3 + x for (i, A) in enumerate(tx) if A == '*'])


def codonfold(sx):
    """Transform a string into a list of codons"""
    assert (len(sx) % 3 == 0)
    return [sx[i:i + 3] for i in range(0, len(sx), 3)]


def isstart(codon):
    """Is this codon a start?"""
    return codon == 'ATG' or codon == 'GTG' or codon == 'TTG'


def findfirststart(sx, m=0):
    """Find the next start codon starting from position m"""
    sfx = codonfold(sx)
    assert(len(sfx) >= m)
    i = m
    while(i < len(sfx)):
        if isstart(sfx[i]):
            return i
        i = i + 1
    return -1


def countstart(sx):
    """How many start codons does the sequence contain?"""
    sfx = codonfold(sx)
    return len([c for c in sfx if isstart(c)])


def removestartinframepx(s0, x, verbose=True):
    """Take a sequence and remove start codons in alternative frame with only synonymous changes in main frame, without creating stop codons
    The sequence is taken as a python string and NOT a BioSeq object
    Returns the modified sequence and the number of remaining starts
    """

    assert(type(s0) == str)
    assert (x > 0) and (x < 3)
    prot0 = Seq(s0).translate()
    l0 = len(s0)
    sx = s0[x:l0 - 3 + x]
    pt = findfirststart(sx)

    nstop = str(Seq(sx).translate()).count('*')

    bummer = False

    changedposition = []

    # While there remains some start codons
    while(pt > -1):
        # Find the involved codons in s0
        p1 = pt * 3
        c1 = s0[p1:p1 + 3]
        c2 = s0[p1 + 3:p1 + 6]
        # Try to change them with synonymous
        candfound = False
        allcand = dict()
        for ((cand1, cand2), rarity_score) in generate_synonymous_combinations([c1, c2]):
            new = list(s0)
            new[p1:p1 + 3] = cand1[0:3]
            new[p1 + 3:p1 + 6] = cand2[0:3]
            news0 = "".join(new)
            # Check whether the start codon has been removed in new sequence
            newsx = news0[x:l0 - 3 + x]
            newpt = findfirststart(newsx, pt)
            newnstop = str(Seq(newsx).translate()).count('*')
            if ((newpt > pt) or (newpt == -1)) and (newnstop <= nstop):  # We removed one start and did not create one stop
                candfound = True
                # Calculate the number of BPS
                allcand[(cand1, cand2)] = ([(c1 + c2)[i] == BP for (i, BP) in enumerate(cand1 + cand2)].count(False), rarity_score)
        if candfound:  # We found at least one candidate
            # Find the best one (less rare codons and less BPS) among all possible candidates
            (cand1, cand2) = sorted(allcand, key=lambda u: allcand[u][0] + int(allcand[u][1] * 10))[0]
            changedposition = changedposition + [p1 + i for (i, BP) in enumerate(cand1 + cand2) if BP != (c1 + c2)[i]]
            new = list(s0)
            new[p1:p1 + 3] = cand1[0:3]
            new[p1 + 3:p1 + 6] = cand2[0:3]
            news0 = "".join(new)
            newsx = news0[x:l0 - 3 + x]
            newpt = findfirststart(newsx, pt)
            newnstop = str(Seq(newsx).translate()).count('*')
            if newnstop < nstop:
                nstop = newnstop
            assert(len(s0) == len(news0))
            if verbose:
                print("We removed start codon at position " + str(pt * 3 + x) + ". ")
            s0 = news0
            sx = newsx
            pt = newpt
        else:
            bummer = True
            if verbose:
                print("Bummer ! We are unable to remove start codon at position " + str(pt * 3 + x) + ". Let's skip this one and go on.")
            pt = findfirststart(sx, pt + 1)
            continue
    # We have removed all start codons in frame x
    assert (str(Seq(s0).translate()) == str(prot0))
    assert (str(Seq(sx).translate()).count('*') <= nstop)  # Actually it is expected to be = and not <
    assert (bummer or findfirststart(sx) == -1)
    return (s0, changedposition, [i * 3 + x for (i, A) in enumerate(codonfold(sx)) if isstart(A)])


def frameshiftability(sequence):
    """Measure the susceptibility for frameshifts due to runs of repeated nucleotides
    The sequence is taken as a python string
    """
    n = 2
    nbrepeats = [0 for nucl in sequence]
    for (i, nucl) in enumerate(sequence):
        if i == 0:
            nbrepeats[i] = 1
            continue
        if (sequence[i-1] == sequence[i]):
            nbrepeats[i] = nbrepeats[i-1]+1
        else:
            nbrepeats[i] = 1
    return nbrepeats


def frameshiftability_score(sequence):
    nbrepeats = frameshiftability(sequence)
    return sum(nbrepeats)-len(sequence)


def removeFShotspots2frame(sequence, frame, maxlrun, begconserve, endconserve):
    """Try to remove runs of 3 base pair or more doing only synonymous changes in main frame, and without creating start and stop codon in alternative frame.
    Input is a python string and not a BioSeq object
    """

    # Check consistency of inputs
    assert(type(sequence) == str)
    assert (frame > 0) and (frame < 3)
    assert(maxlrun > 1)
    assert(begconserve < endconserve)

    # Record initial number of starts and stops in alternative frame (+frame)
    protein = Seq(sequence).translate()
    l0 = len(sequence)
    sx = sequence[frame:l0 - 3 + frame]
    initnbstarts = countstart(sx)
    initnbstops = str(Seq(sx).translate()).count('*')

    # Find all runs
    initsequence = sequence
    nbrepeats = frameshiftability(sequence)
    startrepeats = [i + 1 - maxlrun for (i, x) in enumerate(nbrepeats) if x == maxlrun]

    # Try to eliminate them
    for firstpos in startrepeats:

        # Find end position
        nucl = sequence[firstpos]
        i = firstpos + 1
        while (i <= len(sequence) - 1 and sequence[i] == nucl):
            i = i + 1
        lastpos = i - 1

        # If this run involves the part of the sequence that should be conserved (RBS-start) then we skip it
        if (firstpos - firstpos % 3 - 3 <= endconserve) and (lastpos - lastpos % 3 + 6 >= begconserve):
            continue

        # Find the involved codons in 'main' frame
        codons = [sequence[k:k + 3] for k in range(firstpos - firstpos % 3, lastpos - lastpos % 3 + 1, 3)]

        # Find the preceding and following codons because: they can also be used to avoid creating start/stop in alternative frame, and will be used to avoid creating a new frameshift hotspot.
        prefixe = ''
        suffixe = ''
        if firstpos >= 3:
            prefixe = sequence[firstpos - firstpos % 3 - 3:firstpos - firstpos % 3]
        if lastpos <= len(sequence) - 4:
            suffixe = sequence[lastpos - lastpos % 3 + 3:lastpos - lastpos % 3 + 6]

        # Record the number of start and stop in alternative frame of local sequence before making changes
        localseqbefore = prefixe + reduce(lambda x, y: x + y, codons) + suffixe
        lcontext = len(suffixe) + len(prefixe) + len(codons) * 3
        assert(len(localseqbefore) == lcontext)
        nbstartsbefore = countstart(localseqbefore[frame:lcontext - 3 + frame])
        nbstopsbefore = str(Seq(localseqbefore[frame:lcontext - 3 + frame]).translate()).count('*')

        # Generate all the possible combinations of synonymous codons and the corresponding local sequence
        allcombination = map(lambda combination: (combination[1], reduce(lambda x, y: x + y, combination[0])), generate_synonymous_combinations(codons))

        # Only keep the combinations that do not create start/stop in alternative frame
        allowedcombination = filter(lambda combination: countstart((prefixe + combination[1] + suffixe)[frame:lcontext - 3 + frame]) <= nbstartsbefore and str(Seq((prefixe + combination[1] + suffixe)[frame:lcontext - 3 + frame]).translate()).count('*') <= nbstopsbefore, allcombination)

        # Sort them according to potentiallity for frame shifts and smaller use or rare codons
        bestcombination = min(allowedcombination, key=lambda combination: frameshiftability_score(prefixe + combination[1] + suffixe) * 100 + int(combination[0] * 10.))
        # could be use as a criteria to refine above metric: when equals in frameshiftability, we could choose the combination that minimizes the number of (synonymous) BPS made.
        # print [x==sequence[firstpos-firstpos%3+i] for (i,x) in enumerate(bestcombination)].count(False)

        # Take the best combination and replace appropriate nucleotides in sequence
        l = list(sequence)
        l[firstpos - firstpos % 3:lastpos - lastpos % 3 + 3] = bestcombination[1]
        sequence = str('').join(l)

    # Assert that we only made synonymous changes
    assert(len(sequence) == l0)
    assert(str(Seq(initsequence).translate()) == str(Seq(sequence).translate()))

    # Assert that we did not created start/stop in alternative frame
    finalnbstarts = countstart(sequence[frame:l0 - 3 + frame])
    finalnbstops = str(Seq(sequence[frame:l0 - 3 + frame]).translate()).count('*')
    assert(finalnbstarts <= initnbstarts)  # warning: not tolerant
    assert(finalnbstops <= initnbstops)  # warning: not tolerant

    # return the new sequence and the remaining hotspots
    newnbrepeats = frameshiftability(sequence)
    remaininghotspots = [i + 1 - maxlrun for (i, x) in enumerate(newnbrepeats) if x == maxlrun]
    return(sequence, remaininghotspots)


def removerarecodonsinframepx(sequence, frame, maxlrun, rarethreshold=8., verbose=True):
    """Try to remove rare codons in alternative frame with the following constraints:
    Do not add starts in alternative frame
    Do not add stops in alternative frame
    Do not create fs hotspots
    Do not add rare codons in main frame
    """
    assert(type(sequence) == str)
    assert (frame > 0) and (frame < 3)
    protein = Seq(sequence).translate()
    l0 = len(sequence)
    sx = sequence[frame:l0 - 3 + frame]
    initnbstarts = countstart(sx)
    initnbstops = str(Seq(sx).translate()).count('*')
    initfs = frameshiftability(sequence)
    initfsscore = sum([x for x in initfs if x >= maxlrun])

    changedposition = []
    remaining = []

    # pt=findfirststart(sx)
    codons = [sx[3 * i:3 * i + 3] for i in range(0, int(len(sx) / 3))]
    listerare = [i for i in range(0, int(len(sx) / 3)) if CodonUsage[codons[i]] < rarethreshold]
    # sumscore=sum[CodonUsage[codons[i]] for i in listrare]
    nbrare = len(listerare)
    while(nbrare > 0):  # for irare in listerare:
        # print ' '
        irare = listerare[0]
        c1 = sequence[3 * irare:3 * irare + 3]
        c2 = sequence[3 * irare + 3:3 * irare + 6]
        # print codons[irare] + ' is rare ' + c1 + ' ' + c2
        candfound = False
        old_rarity_score = compute_rarity_score([c1, c2])  # Rarity score of the six bp sequence in main frame
        old_usage_framepx = CodonUsage[codons[irare]]  # Codon usage of the involved rare codon in alternative frame
        allcand = dict()
        for ((cand1, cand2), rarity_score) in generate_synonymous_combinations([c1, c2]):  # Warning: rarity_score is calculated in the frame of GalK.
            # print cand1 + ' ' + cand2 + ' ' + str(irare)
            new = list(sequence)
            new[3 * irare:3 * irare + 3] = cand1[0:3]
            new[3 * irare + 3:3 * irare + 6] = cand2[0:3]
            newsequence = "".join(new)
            newsx = newsequence[frame:l0 - 3 + frame]
            # print newsequence + ' ' + newsx
            # Check that this candidate does not add start/stop codons or fs hotspots or rare codons in galK frame or rare codon upstream in alternative frame
            if countstart(newsx) > initnbstarts:
                # print 'no (start)'
                continue
            if str(Seq(newsx).translate()).count('*') > initnbstops:
                # print 'no (stop)'
                continue
            newfs = frameshiftability(newsequence)
            newfsscore = sum([x for x in newfs if x >= maxlrun])
            if newfsscore > initfsscore:
                # print 'no (fs)'
                continue
            if rarity_score > old_rarity_score:
                # print 'no (rare in main frame)'
                continue
            newcodons = [newsx[3 * i:3 * i + 3] for i in range(0, int(len(newsx) / 3))]
            newlisterare = [i for i in range(0, int(len(newsx) / 3)) if CodonUsage[newcodons[i]] < rarethreshold]
            diff = set(newlisterare) - set(remaining)
            if len(diff) > 0 and min(diff) < irare:  # Are we adding rare codons in candidate before the one we are trying to remove?
                # print 'no (rare in alt frame)'
                continue
            # This candidate is valid
            candfound = True
            altframecodon = newsx[3 * irare:3 * irare + 3]
            # Record the number of BPS and the rarity score in alternative frame
            allcand[(cand1, cand2)] = ([(c1 + c2)[i] == BP for (i, BP) in enumerate(cand1 + cand2)].count(False), CodonUsage[altframecodon])

        if candfound:  # We found at least one candidate
            # Find the best one (less BPS and less rare codons) among all possible candidates
            sortedcandidates = sorted(allcand, key=lambda u: allcand[u][0] - int(allcand[u][1] * 10.))
            (cand1, cand2) = sortedcandidates[0]
            new_usage_framepx = allcand[sortedcandidates[0]][1]
            new = list(sequence)
            new[3 * irare:3 * irare + 3] = cand1[0:3]
            new[3 * irare + 3:3 * irare + 6] = cand2[0:3]
            newsequence = "".join(new)
            newsx = newsequence[frame:l0 - 3 + frame]
            assert(len(sequence) == len(newsequence))
            assert (str(Seq(sequence).translate()) == str(Seq(newsequence).translate()))

            newcodons = [newsx[3 * i:3 * i + 3] for i in range(0, int(len(newsx) / 3))]
            newlisterare = set([i for i in range(0, int(len(newsx) / 3)) if CodonUsage[newcodons[i]] < rarethreshold]) - set(remaining)

            # Is this candidate better than original sequence ?
            if new_usage_framepx > old_usage_framepx:
                sequence = newsequence
                codons = newcodons
                # listerare.remove(irare)
                listerare = sorted(list(newlisterare))
                nbrare = len(listerare)
                changedposition = changedposition + [3 * irare + i for (i, BP) in enumerate(cand1 + cand2) if BP != (c1 + c2)[i]]
                # print sequence + ' is new sequence'
                continue

        # No valid candidate found
        remaining = remaining + [irare]
        listerare.remove(irare)
        nbrare = len(listerare)
        if verbose:
            print("unable to remove rare codon " + codons[irare] + " at position " + str(3 * irare + frame))
    return (sequence, changedposition, [3 * x + frame for x in remaining])


